import pytest

from bluetooth_mesh.messages.generic.property import (
    AdminUserAccess,
    GenericPropertyMessage,
    GenericPropertyOpcode,
    ManufacturerUserAccess,
    UserAccess,
)
from bluetooth_mesh.messages.properties import LightDistributionField, LightSourceTypeField, PropertyID

valid = [
    # fmt: off
    pytest.param(
        b'\x82\x2E',
        GenericPropertyOpcode.GENERIC_USER_PROPERTIES_GET,
        dict(),
        id="GENERIC_USER_PROPERTIES_GET"),
    pytest.param(
        b'\x4B\x8C\x00\x04\x00\x09\x00',
        GenericPropertyOpcode.GENERIC_USER_PROPERTIES_STATUS,
        dict(property_ids=[PropertyID.LIGHT_DISTRIBUTION,
                           PropertyID.AVERAGE_OUTPUT_CURRENT,
                           PropertyID.COLOR_RENDERING_INDEX_RA]),
        id="GENERIC_USER_PROPERTIES_STATUS"),
    pytest.param(
        b'\x82\x2F\x8C\x00',
        GenericPropertyOpcode.GENERIC_USER_PROPERTY_GET,
        dict(property_id=PropertyID.LIGHT_DISTRIBUTION),
        id="GENERIC_USER_PROPERTY_GET"),
    pytest.param(
        b'\x4C\x8C\x00\x05',
        GenericPropertyOpcode.GENERIC_USER_PROPERTY_SET,
        dict(property_id=PropertyID.LIGHT_DISTRIBUTION,
             light_distribution=dict(
                 light_distribution=LightDistributionField.TYPE_V)),
        id="GENERIC_USER_PROPERTY_SET"),
    pytest.param(
        b'\x4D\x8C\x00\x05',
        GenericPropertyOpcode.GENERIC_USER_PROPERTY_SET_UNACKNOWLEDGED,
        dict(property_id=PropertyID.LIGHT_DISTRIBUTION,
             light_distribution=dict(
                 light_distribution=LightDistributionField.TYPE_V)),
        id="GENERIC_USER_PROPERTY_SET_UNACKNOWLEDGED"),
    pytest.param(
        b'\x4E\x8C\x00\x01\x05',
        GenericPropertyOpcode.GENERIC_USER_PROPERTY_STATUS,
        dict(property_id=PropertyID.LIGHT_DISTRIBUTION,
             access=UserAccess.READ_ONLY,
             light_distribution=dict(
                 light_distribution=LightDistributionField.TYPE_V)),
        id="GENERIC_USER_PROPERTY_STATUS"),
    pytest.param(
        b'\x82\x2C',
        GenericPropertyOpcode.GENERIC_ADMIN_PROPERTIES_GET,
        dict(),
        id="GENERIC_ADMIN_PROPERTIES_GET"),
    pytest.param(
        b'\x47\x8C\x00\xA4\x00\xB6\x00\xAB\x00',
        GenericPropertyOpcode.GENERIC_ADMIN_PROPERTIES_STATUS,
        dict(property_ids=[PropertyID.LIGHT_DISTRIBUTION,
                           PropertyID.NOMINAL_LIGHT_OUTPUT,
                           PropertyID.THERMAL_DERATING,
                           PropertyID.RATED_MEDIAN_USEFUL_LIFE_OF_LUMINAIRE]
             ),
        id="GENERIC_ADMIN_PROPERTIES_STATUS"),
    pytest.param(
        b'\x82\x2D\x8C\x00',
        GenericPropertyOpcode.GENERIC_ADMIN_PROPERTY_GET,
        dict(property_id=PropertyID.LIGHT_DISTRIBUTION),
        id="GENERIC_ADMIN_PROPERTY_GET"),
    pytest.param(
        b'\x48\x94\x00\x03\x01\x80',
        GenericPropertyOpcode.GENERIC_ADMIN_PROPERTY_SET,
        dict(property_id=PropertyID.LIGHT_SOURCE_TEMPERATURE,
             access=AdminUserAccess.READ_WRITE,
             light_source_temperature=dict(
                 temperature=-16383.5)),  # It supposed to be value not know
        id="GENERIC_ADMIN_PROPERTY_SET"),
    pytest.param(
        b'\x49\xA4\x00\x01\xFF\xFF\xFF',
        GenericPropertyOpcode.GENERIC_ADMIN_PROPERTY_SET_UNACKNOWLEDGED,
        dict(property_id=PropertyID.NOMINAL_LIGHT_OUTPUT,
             access=AdminUserAccess.READ_ONLY,
             nominal_light_output=dict(
                 light_output=16777215)), # It supposed to be value not know
        id="GENERIC_ADMIN_PROPERTY_SET_UNACKNOWLEDGED"),
    pytest.param(
        b'\x4A\xA4\x00\x01\x0A\x00\x00',
        GenericPropertyOpcode.GENERIC_ADMIN_PROPERTY_STATUS,
        dict(property_id=PropertyID.NOMINAL_LIGHT_OUTPUT,
             access=UserAccess.READ_ONLY,
             nominal_light_output=dict(
                 light_output=10)),
        id="GENERIC_ADMIN_PROPERTY_STATUS"),
    pytest.param(
        b'\x82\x2A',
        GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTIES_GET,
        dict(),
        id="GENERIC_MANUFACTURER_PROPERTIES_GET"),
    pytest.param(
        b'\x43\x9d\x00\x8F\x00\xB6\x00\x07\x00',
        GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTIES_STATUS,
        dict(property_ids=[PropertyID.LUMINAIRE_NOMINAL_MAXIMUM_AC_MAINS_VOLTAGE,
                           PropertyID.LIGHT_SOURCE_ON_TIME_RESETTABLE,
                           PropertyID.THERMAL_DERATING,
                           PropertyID.CHROMATICITY_TOLERANCE]
             ),
        id="GENERIC_MANUFACTURER_PROPERTIES_STATUS"),
    pytest.param(
        b'\x82\x2D\xA0\x00',
        GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_GET,
        dict(property_id=PropertyID.LUMINAIRE_TIME_OF_MANUFACTURE),
        id="GENERIC_MANUFACTURER_PROPERTY_GET"),
    pytest.param(
        b'\x44\x73\x00\x01',
        GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_SET,
        dict(property_id=PropertyID.POWER_FACTOR,
             access=ManufacturerUserAccess.READ_ONLY),
        id="GENERIC_MANUFACTURER_PROPERTY_SET"),
    pytest.param(
        b'\x45\x89\x00\x00',
        GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_SET_UNACKNOWLEDGED,
        dict(property_id=PropertyID.EXTERNAL_SUPPLY_VOLTAGE_FREQUENCY,
             access=ManufacturerUserAccess.NOT_USER_PROPERTY),
        id="GENERIC_MANUFACTURER_PROPERTY_SET_UNACKNOWLEDGED"),
    pytest.param(
        b'\x46\x9A\x00\x01abcdefghijklmnoprstuvxyz',
        GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_STATUS,
        dict(property_id=PropertyID.LUMINAIRE_IDENTIFICATION_NUMBER,
             access=UserAccess.READ_ONLY,
             luminaire_identification_number="abcdefghijklmnoprstuvxyz"),
        id="GENERIC_MANUFACTURER_PROPERTY_STATUS"),
    pytest.param(
        b'\x46\xB3\x00\x01\x03',
        GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_STATUS,
        dict(property_id=PropertyID.LIGHT_SOURCE_TYPE,
             access=UserAccess.READ_ONLY,
             light_source_type=dict(light_source_type=
                 LightSourceTypeField.LOW_VOLTAGE_HALOGEN)),
        id="GENERIC_MANUFACTURER_PROPERTY_STATUS_LIGHT_SOURCE"),
    pytest.param(
        b'\x4F',
        GenericPropertyOpcode.GENERIC_CLIENT_PROPERTIES_GET,
        dict(),
        id="GENERIC_CLIENT_PROPERTIES_GET"),
    pytest.param(
        b'\x50\x9d\x00\x8F\x00\xB6\x00\x07\x00',
        GenericPropertyOpcode.GENERIC_CLIENT_PROPERTIES_STATUS,
        dict(property_ids=[PropertyID.LUMINAIRE_NOMINAL_MAXIMUM_AC_MAINS_VOLTAGE,
                           PropertyID.LIGHT_SOURCE_ON_TIME_RESETTABLE,
                           PropertyID.THERMAL_DERATING,
                           PropertyID.CHROMATICITY_TOLERANCE]
             ),
        id="GENERIC_MANUFACTURER_PROPERTIES_STATUS"),
]


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_parse_valid(encoded, opcode, data):
    assert GenericPropertyMessage.parse(encoded).params == data



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2023  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from datetime import datetime, timedelta, timezone
from enum import IntEnum

from construct import (
    Adapter,
    BitsInteger,
    BitStruct,
    BytesInteger,
    Const,
    Construct,
    Container,
    Embedded,
    Flag,
    Int8ul,
    Int16sl,
    Int16ul,
    Int32ul,
    Padding,
    Struct,
    this,
)

from bluetooth_mesh.messages.properties import PropertyDict, PropertyMixin, TimeSecond32
from bluetooth_mesh.messages.time import (
    MESH_UNIX_EPOCH_DIFF,
    DateTime,
    TAIUTCDeltaAdapter,
    mesh_tai_utc_delta_to_timedelta,
    mesh_time_zone_offset_to_timedelta,
    timedelta_to_mesh_tai_utc_delta,
    timedelta_to_mesh_time_zone_offset,
)
from bluetooth_mesh.messages.util import (
    AliasedContainer,
    EmbeddedBitStruct,
    EnumAdapter,
    EnumSwitch as Switch,
    NamedSelect,
    Opcode,
    SwitchStruct,
)


class EmergencyLightingTestOpcode(IntEnum):
    SILVAIR_ELT = 0xE93601


class EmergencyLightingTestSubOpcode(IntEnum):
    ELT_FUNCTIONAL_TEST_GET = 0x00
    ELT_FUNCTIONAL_TEST_START = 0x01
    ELT_FUNCTIONAL_TEST_STOP = 0x02
    ELT_FUNCTIONAL_TEST_STATUS = 0x03
    ELT_DURATION_TEST_GET = 0x04
    ELT_DURATION_TEST_START = 0x05
    ELT_DURATION_TEST_STOP = 0x06
    ELT_DURATION_TEST_STATUS = 0x07
    ELT_PROPERTY_GET = 0x08
    ELT_PROPERTY_SET = 0x09
    ELT_PROPERTY_SET_UNACKNOWLEDGED = 0x0A
    ELT_PROPERTY_STATUS = 0x0B


class TestExecutionStatus(IntEnum):
    FINISHED = 0
    TRIGGERED = 1
    IN_PROGRESS = 2
    POSTPONED = 3
    STOPPED = 4
    DROPPED = 5
    TIMEOUT = 6
    UNKNOWN = 7


FunctionalExecutionResult = BitStruct(
    Padding(5),
    "circuit_fault" / Flag,
    "battery_fault" / Flag,
    "lamp_fault" / Flag,
)

DurationExecutionResult = BitStruct(
    Padding(4),
    "battery_duration_fault" / Flag,
    "circuit_fault" / Flag,
    "battery_fault" / Flag,
    "lamp_fault" / Flag,
)


class TimestampAdapter(Adapter):
    _subcon = Struct(
        "date" / DateTime,
        "tai_utc_delta" / TAIUTCDeltaAdapter(Int16sl),
    )

    def _decode(self, obj, context, path):
        tai_utc_delta_raw = obj["tai_utc_delta_raw"]
        tai_utc_delta_padding = (tai_utc_delta_raw >> 15) & 0x1
        tai_utc_delta_encoded = tai_utc_delta_raw & 0x7FFF
        tai_utc_delta = mesh_tai_utc_delta_to_timedelta(tai_utc_delta_encoded)

        if tai_utc_delta_padding != 0:
            raise ValueError("Invalid TAI-UTC delta: padding bit must be 0")

        time_zone_offset = mesh_time_zone_offset_to_timedelta(obj["time_zone_offset"])
        full_recv_time = obj["tai_seconds"] + MESH_UNIX_EPOCH_DIFF
        recv_date = datetime.fromtimestamp(full_recv_time, timezone(time_zone_offset)) - tai_utc_delta

        return Container(
            date=recv_date,
            tai_utc_delta=tai_utc_delta,
        )

    def _encode(self, obj, context, path):
        passed_time: datetime = obj["date"]

        if isinstance(passed_time, dict):  # capnproto message
            time_zone = timedelta(minutes=passed_time["time_zone_offset"])
            passed_time = datetime(
                year=passed_time["year"],
                month=passed_time["month"],
                day=passed_time["day"],
                hour=passed_time["hour"],
                minute=passed_time["minute"],
                second=passed_time["second"],
                microsecond=passed_time["microsecond"],
                tzinfo=timezone(time_zone),
            )

        if isinstance(obj["tai_utc_delta"], int):
            obj["tai_utc_delta"] = timedelta(seconds=obj["tai_utc_delta"])

        tai_utc_delta = obj["tai_utc_delta"]
        tz_offset = passed_time.utcoffset()

        utc_time = passed_time.astimezone(timezone.utc)
        tai_time = utc_time + tai_utc_delta
        tai_seconds = int(tai_time.timestamp()) - MESH_UNIX_EPOCH_DIFF

        return Container(
            tai_seconds=tai_seconds,
            tai_utc_delta_raw=timedelta_to_mesh_tai_utc_delta(tai_utc_delta),
            time_zone_offset=timedelta_to_mesh_time_zone_offset(tz_offset),
        )


class EmergencyLightingTestProperty(IntEnum):
    ELT_DURATION_TEST_EXECUTION_TIMEOUT = 0xFF84
    ELT_FUNCTIONAL_TEST_EXECUTION_TIMEOUT = 0xFF85
    ELT_DURATION_TEST_RETRY_PERIOD = 0xFF86
    ELT_FUNCTIONAL_TEST_RETRY_PERIOD = 0xFF87
    ELT_DURATION_TEST_BACKUP_AUTOMATIC_DELAY = 0xFF88
    ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_DELAY = 0xFF89
    ELT_DURATION_TEST_BACKUP_AUTOMATIC_INTERVAL = 0xFF8A
    ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_INTERVAL = 0xFF8B


EmergencyLightingTestPropertyDict = {
    EmergencyLightingTestProperty.ELT_DURATION_TEST_EXECUTION_TIMEOUT: TimeSecond32,
    EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_EXECUTION_TIMEOUT: TimeSecond32,
    EmergencyLightingTestProperty.ELT_DURATION_TEST_RETRY_PERIOD: TimeSecond32,
    EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_RETRY_PERIOD: TimeSecond32,
    EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_DELAY: TimeSecond32,
    EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_DELAY: TimeSecond32,
    EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_INTERVAL: TimeSecond32,
    EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_INTERVAL: TimeSecond32,
}


class _EmergencyLightingTestProperty(PropertyMixin, Construct):
    ENUM = EmergencyLightingTestProperty
    DICT = EmergencyLightingTestPropertyDict

    subcon = Struct(
        "property_id" / EnumAdapter(Int16ul, EmergencyLightingTestProperty),
        Switch(this.id, EmergencyLightingTestPropertyDict),
    )

    def _parse(self, stream, context, path):
        msg = EmergencyLightingTestPropertyGet._parse(stream, context, path)
        return self._parse_property(msg, stream, context, path)

    def _build(self, obj, stream, context, path):
        EmergencyLightingTestPropertyGet._build(obj, stream, context, path)
        return self._build_property(obj, stream, context, path)


# fmt: off
Timestamp = Struct(
    "tai_seconds" / BytesInteger(5, swapped=True),
    "time_zone_offset" / Int8ul,
    "tai_utc_delta_raw" / BytesInteger(2, swapped=True),
)

EmergencyLightingTestFunctionalTestGet = Struct()

EmergencyLightingTestFunctionalTestStart = Struct()

EmergencyLightingTestFunctionalTestStop = Struct()

EmergencyLightingTestFunctionalTestStatusMinimal = Struct(
    "tai_timestamp" / TimestampAdapter(Timestamp),
    "execution_status" / EnumAdapter(Int8ul, TestExecutionStatus),
    "execution_result" / FunctionalExecutionResult,
)

EmergencyLightingTestFunctionalTestStatusOptional = Struct(
    Embedded(EmergencyLightingTestFunctionalTestStatusMinimal),
    "relative_timestamp" / Int32ul,
)

EmergencyLightingTestFunctionalTestStatus = NamedSelect(
    optional=EmergencyLightingTestFunctionalTestStatusOptional,
    minimal=EmergencyLightingTestFunctionalTestStatusMinimal,
)

EmergencyLightingTestDurationTestGet = Struct()

EmergencyLightingTestDurationTestStart = Struct()

EmergencyLightingTestDurationTestStop = Struct()

EmergencyLightingTestDurationTestStatusMinimal = Struct(
    "tai_timestamp" / TimestampAdapter(Timestamp),
    "execution_status" / EnumAdapter(Int8ul, TestExecutionStatus),
    "execution_result" / DurationExecutionResult,
    "duration_result" / Int16ul,
)

EmergencyLightingTestDurationTestStatusOptional = Struct(
    Embedded(EmergencyLightingTestDurationTestStatusMinimal),
    "relative_timestamp" / Int32ul,
)

EmergencyLightingTestDurationTestStatus = NamedSelect(
    optional=EmergencyLightingTestDurationTestStatusOptional,
    minimal=EmergencyLightingTestDurationTestStatusMinimal,
)


EmergencyLightingTestPropertyGet = Struct(
    "property_id" / EnumAdapter(Int16ul, EmergencyLightingTestProperty),
)

EmergencyLightingTestPropertySet = _EmergencyLightingTestProperty()

EmergencyLightingTestPropertyStatus = _EmergencyLightingTestProperty()


EmergencyLightingTestParams = SwitchStruct(
    "subopcode" / EnumAdapter(Int8ul, EmergencyLightingTestSubOpcode),
    "payload" / Switch(
        this.subopcode,
        {
            EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_GET: EmergencyLightingTestFunctionalTestGet,
            EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_START: EmergencyLightingTestFunctionalTestStart,
            EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_STOP: EmergencyLightingTestFunctionalTestStop,
            EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_STATUS: EmergencyLightingTestFunctionalTestStatus,
            EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_GET: EmergencyLightingTestDurationTestGet,
            EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_START: EmergencyLightingTestDurationTestStart,
            EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_STOP: EmergencyLightingTestDurationTestStop,
            EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_STATUS: EmergencyLightingTestDurationTestStatus,
            EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET: EmergencyLightingTestPropertyGet,
            EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET: EmergencyLightingTestPropertySet,
            EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED: EmergencyLightingTestPropertySet,
            EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS: EmergencyLightingTestPropertyStatus,
        }
    )
)

EmergencyLightingTestMessage = SwitchStruct(
    "opcode" / Opcode(EmergencyLightingTestOpcode),
    "params" / Switch(
        this.opcode,
        {
            EmergencyLightingTestOpcode.SILVAIR_ELT: EmergencyLightingTestParams
        }
    )
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import importlib
import json
import logging
from tempfile import NamedTemporaryFile

import pytest

from bluetooth_mesh.messages import AccessMessage
from bluetooth_mesh.messages.util import to_camelcase_dict, to_snakecase_dict

valid = [
    # Silvair Debug Server
    bytes.fromhex("F53601" + "00"),  # RSSI_THRESHOLD_GET
    bytes.fromhex("F53601" + "01" + "f0"),  # RSSI_THRESHOLD_SET
    bytes.fromhex("F53601" + "02" + "f0"),  # RSSI_THRESHOLD_STATUS
    bytes.fromhex("F53601" + "03" + "ff"),  # RADIO_TEST
    bytes.fromhex("F53601" + "0a"),  # UPTIME_GET
    bytes.fromhex("F53601" + "0b" + "00000000"),  # UPTIME_STATUS
    bytes.fromhex("F53601" + "0b" + "ffffffff"),  # UPTIME_STATUS
    bytes.fromhex("F53601" + "0c"),  # LAST_SW_FAULT_GET
    bytes.fromhex("F53601" + "0d"),  # LAST_SW_FAULT_CLEAR
    bytes.fromhex("F53601" + "0e" + "00000000" + "616263"),  # LAST_SW_FAULT_STATUS
    bytes.fromhex("F53601" + "0f"),  # SYSTEM_STATS_GET
    bytes.fromhex("F53601" + "10"),  # SYSTEM_STATS_STATUS
    bytes.fromhex("F53601" + "10" + "6162636465666768000000000000"),
    bytes.fromhex("F53601" + "10" + "6162636400000000ffff00000000"),
    bytes.fromhex("F53601" + "11"),  # LAST_MALLOC_FAULT_GET
    bytes.fromhex("F53601" + "12"),  # LAST_MALLOC_FAULT_CLEAR
    bytes.fromhex("F53601" + "13" + "00000000" + "616263"),  # LAST_MALLOC_FAULT_STATUS
    bytes.fromhex("F53601" + "14"),  # LAST_FDS_FAULT_GET
    bytes.fromhex("F53601" + "15"),  # LAST_FDS_FAULT_CLEAR
    bytes.fromhex("F53601" + "16" + "00000000" + "616263"),  # LAST_FDS_FAULT_STATUS
    bytes.fromhex("F53601" + "17"),  # BYTES_BEFORE_GARBAGE_COLLECTOR_GET
    bytes.fromhex("F53601" + "18" + "0000"),  # BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS
    bytes.fromhex("F53601" + "18" + "1234"),  # BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS
    bytes.fromhex("F53601" + "18" + "ffff"),  # BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS
    bytes.fromhex("F53601" + "19"),  # PROVISIONED_APP_VERSION_GET
    bytes.fromhex("F53601" + "1a" + "0000"),  # PROVISIONED_APP_VERSION_STATUS
    bytes.fromhex("F53601" + "1a" + "1234"),  # PROVISIONED_APP_VERSION_STATUS
    bytes.fromhex("F53601" + "1a" + "ffff"),  # PROVISIONED_APP_VERSION_STATUS
    bytes.fromhex("F53601" + "1b"),  # FULL_FIRMWARE_VERSION_GET
    bytes.fromhex("F53601" + "1c" + "61626364"),  # FULL_FIRMWARE_VERSION_STATUS
    bytes.fromhex("F53601" + "1d"),  # IV_INDEX_GET
    bytes.fromhex("F53601" + "1e" + "00000000"),  # IV_INDEX_STATUS
    bytes.fromhex("F53601" + "1e" + "12345678"),  # IV_INDEX_STATUS
    bytes.fromhex("F53601" + "1e" + "ffffffff"),  # IV_INDEX_STATUS
    bytes.fromhex("F53601" + "1f"),  # GARBAGE_COLLECTOR_COUNTER_GET
    bytes.fromhex("F53601" + "20" + "0000"),  # GARBAGE_COLLECTOR_COUNTER_STATUS
    bytes.fromhex("F53601" + "20" + "1234"),  # GARBAGE_COLLECTOR_COUNTER_STATUS
    bytes.fromhex("F53601" + "20" + "ffff"),  # GARBAGE_COLLECTOR_COUNTER_STATUS
    bytes.fromhex("F53601" + "21"),  # ARAP_LIST_SIZE_GET
    bytes.fromhex("F53601" + "22" + "0000"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("F53601" + "22" + "1234"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("F53601" + "22" + "ffff"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("F53601" + "22" + "00000000"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("F53601" + "22" + "12345678"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("F53601" + "22" + "ffffffff"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("F53601" + "23" + "00"),  # ARAP_LIST_CONTENT_GET
    bytes.fromhex("F53601" + "23" + "12"),  # ARAP_LIST_CONTENT_GET
    bytes.fromhex("F53601" + "23" + "ff"),  # ARAP_LIST_CONTENT_GET
    bytes.fromhex("F53601" + "24" + "0000"),  # ARAP_LIST_CONTENT_STATUS
    bytes.fromhex("F53601" + "24" + "00" + "00" + "0000000000"),  # ARAP_LIST_CONTENT_STATUS
    bytes.fromhex("F53601" + "24" + "00" + "00" + "ffffffffff"),  # ARAP_LIST_CONTENT_STATUS
    bytes.fromhex(
        "F53601" + "24" + "00" + "00" + "ffffffffff00000000001234561234"
    ),  # ARAP_LIST_CONTENT_STATUS
    # ------------------
    # Configuration Client
    bytes.fromhex("02" + "00" + "3601CE00FECAEFBE0BB000000000"),  # CONFIG_COMPOSITION_DATA_STATUS
    bytes.fromhex("8002" + "00" + "0b00010000012100"),  # CONFIG_APPKEY_LIST
    bytes.fromhex("8039" + "01" + "02" + "03" + "04" + "0506070809"),  # CONFIG_HEARBEAT_PUBLICATION_SET
    bytes.fromhex("803E" + "00" + "A70202000213"),  # CONFIG_MODEL_APP_STATUS
    # ------------------
    # Light CTL Server
    bytes.fromhex("8265" + "2222" + "3333" + "ff" + "32" + "3c"),  # LIGHT_CTL_TEMPERATURE_SET_UNACKNOWLEDGED
    # ------------------
    # Network Diagnostic Setup Server
    bytes.fromhex("fd3601" + "00"),  # PUBLICATION_GET
    bytes.fromhex("fd3601" + "01" + "1234aaaa82031000"),  # PUBLICATION_SET
    bytes.fromhex("fd3601" + "01" + "1234aaaa820310000300"),  # PUBLICATION_SET
    bytes.fromhex("fd3601" + "01" + "1234aaaa82031000"),  # PUBLICATION_SET
    bytes.fromhex("fd3601" + "02" + "1234aaaa820310000300"),  # PUBLICATION_STATUS
    bytes.fromhex("fc3601" + "00"),  # SUBSCRIPTION_GET
    bytes.fromhex("fc3601" + "01" + "0123aaaa"),  # SUBSCRIPTION_SET
    bytes.fromhex("fc3601" + "02" + "0123aaaa"),  # SUBSCRIPTION_SET_UNACKNOWLEDGED
    bytes.fromhex("fc3601" + "03" + "1234aaaa20012300110101"),  # SUBSCRIPTION_STATUS
    # ------------------
    # Sensor Server
    bytes.fromhex("8230"),  # SENSOR_DESCRIPTOR_GET
    bytes.fromhex("8230" + "0400"),  # SENSOR_DESCRIPTOR_GET
    bytes.fromhex("8231"),  # SENSOR_GET
    bytes.fromhex("8231" + "0700"),  # SENSOR_GET
    bytes.fromhex("51" + "0c00000000040b0c"),  # SENSOR_DESCRIPTOR_STATUS
    bytes.fromhex("51" + "1900"),  # SENSOR_DESCRIPTOR_STATUS
    bytes.fromhex("51" + "0c00000000020b0c1f00efcdab071b1c"),  # SENSOR_DESCRIPTOR_STATUS
    bytes.fromhex("52" + "e20ac800"),  # SENSOR_STATUS
    bytes.fromhex("52" + "220b2003"),  # SENSOR_STATUS
    bytes.fromhex("52" + "440da244ff"),  # SENSOR_STATUS
    bytes.fromhex("52" + "44" + "0da2" + "44ff" + "220b2003"),  # SENSOR_STATUS
    bytes.fromhex("52" + "09" + "9040" + "a244" + "ff0000"),  # SENSOR_STATUS
    bytes.fromhex("58" + "3000" + "0100" + "0400" + "0900"),  # SENSOR_SETTINGS_STATUS
    bytes.fromhex("58" + "3000"),  # SENSOR_SETTINGS_STATUS
    bytes.fromhex("59" + "5700" + "5700" + "c800"),  # SENSOR_SETTING_SET
    bytes.fromhex("5b" + "5700" + "5700" + "01" + "c800"),  # SENSOR_SETTING_STATUS
    bytes.fromhex("5b" + "5700" + "0200" + "01" + "c80039"),  # SENSOR_SETTING_STATUS
    bytes.fromhex("59" + "5700" + "5700" + "c800"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "5700" + "0200" + "c80039"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "0500" + "0500" + "200354"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "5700" + "0200" + "c80000"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "5900" + "5900" + "0003"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "4200" + "4d00" + "01"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "4200" + "4200" + "50"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "0a00" + "3600" + "b80b00"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "6d00" + "6d00" + "0a0000"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "6d00" + "6d00" + "ffffff"),
    bytes.fromhex("59" + "550055001a2700"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "4c004c001b1a"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "6c006c00ff1b1a"),  # SENSOR_SETTING_SET
    bytes.fromhex("59" + "6c00" + "6c00" + "ffffff"),  # SENSOR_SETTING_SET + TOTAL_DEVICE_POWER_ON_CYCLES
    bytes.fromhex("59" + "6800" + "6800" + "0500"),  # SENSOR_SETTING_SET + TIME_SINCE_MOTION_SENSED
    bytes.fromhex(
        "59" + "6700" + "6700" + "050001007040"
    ),  # SENSOR_SETTING_SET + SHORT_CIRCUIT_EVENT_STATISTICS
    bytes.fromhex(
        "59" + "0e00" + "0e00" + "6162636465666768"
    ),  # SENSOR_SETTING_SET + DEVICE_FIRMWARE_REVISION
    bytes.fromhex(
        "59" + "0e00" + "0e00" + "6162636465660000"
    ),  # SENSOR_SETTING_SET + DEVICE_FIRMWARE_REVISION
    bytes.fromhex(
        "59" + "1100" + "1100" + "616263646566676861626364656667686162636465666768616263646566676861626364"
    ),
    # SENSOR_SETTING_SET + DEVICE_MANUFACTURER_NAME
    bytes.fromhex("59" + "6a00" + "6a00" + "a244ff"),  # SENSOR_SETTING_SET + TOTAL_DEVICE_ENERGY_USE
    bytes.fromhex("59" + "2e00" + "2e00" + "44ff"),  # SENSOR_SETTING_SET + LIGHT_CONTROL_LIGHTNESS_ON
    bytes.fromhex("59" + "3200" + "3200" + "44ff0000"),  # SENSOR_SETTING_SET + LIGHT_CONTROL_REGULATOR_KID
    bytes.fromhex("59" + "5200" + "5200" + "a08601"),  # SENSOR_SETTING_SET + PRESENT_DEVICE_INPUT_POWER
    bytes.fromhex("59" + "1600" + "1600" + "e80300d007000f2700"),
    # SENSOR_SETTING_SET + DEVICE_POWER_RANGE_SPECIFICATION
    bytes.fromhex("59" + "4f00" + "4f00" + "1f"),  # SENSOR_SETTING_SET + PRESENT_AMBIENT_TEMPERATURE
    bytes.fromhex("59" + "4f00" + "4f00" + "e1"),  # SENSOR_SETTING_SET + PRESENT_AMBIENT_TEMPERATURE
    bytes.fromhex(
        "59" + "5400" + "5400" + "e620"
    ),  # SENSOR_SETTING_SET + PRESENT_DEVICE_OPERATING_TEMPERATURE
    bytes.fromhex("59" + "4500" + "4500" + "3102dc6e71"),  # SENSOR_SETTING_SET + OUTDOOR_STATISTICAL_VALUES
    bytes.fromhex("59" + "1400" + "1400" + "92096400f8f8204e71"),
    # SENSOR_SETTING_SET + DEVICE_OPERATING_TEMPERATURE_STATISTICAL_VALUES
    bytes.fromhex("59" + "1300" + "1300" + "dc6e"),
    # SENSOR_SETTING_SET + DEVICE_OPERATING_TEMPERATURE_RANGE_SPECIFICATION
    bytes.fromhex("59" + "0100" + "0100" + "186da2"),
    # SENSOR_SETTING_SET + AVERAGE_AMBIENT_TEMPERATURE_IN_A_PERIOD_OF_DAY
    bytes.fromhex("59" + "6000" + "6000" + "a244ff6da2"),
    # SENSOR_SETTING_SET + RELATIVE_DEVICE_ENERGY_USE_IN_A_PERIOD_OF_DAY
    bytes.fromhex(
        "59" + "2a00" + "2a00" + "690140000000f0ff54"
    ),  # SENSOR_SETTING_SET + INPUT_VOLTAGE_STATISTICS
    bytes.fromhex("59" + "4900" + "4900" + "00000006f0ff"),  # SENSOR_SETTING_SET + OUTPUT_VOLTAGE_RANGE
    bytes.fromhex(
        "59" + "4700" + "4700" + "9709640000007d1571"
    ),  # SENSOR_SETTING_SET + OUTPUT_CURRENT_STATISTICS
    bytes.fromhex("59" + "4600" + "4600" + "00007d15"),  # SENSOR_SETTING_SET + OUTPUT_CURRENT_RANGE
    bytes.fromhex(
        "59" + "2100" + "2100" + "000001007d15"
    ),  # SENSOR_SETTING_SET + INPUT_CURRENT_RANGE_SPECIFICATION
    bytes.fromhex("59" + "7000" + "7000" + "020000"),  # SENSOR_SETTING_SET + TOTAL_LUMINOUS_ENERGY
    bytes.fromhex(
        "59" + "0600" + "0600" + "0020"
    ),  # SENSOR_SETTING_SET + CENTER_BEAM_INTENSITY_AT_FULL_POWER
    bytes.fromhex("59" + "4000" + "4000" + "020000"),  # SENSOR_SETTING_SET + LUMINOUS_EXPOSURE
    bytes.fromhex("59" + "1f00" + "1f00" + "d007"),  # SENSOR_SETTING_SET + INITIAL_LUMINOUS_FLUX
    bytes.fromhex("59" + "4100" + "4100" + "e803d007"),  # SENSOR_SETTING_SET + LUMINOUS_FLUX_RANGE
    bytes.fromhex("59" + "3e00" + "3e00" + "d407"),  # SENSOR_SETTING_SET + LUMINOUS_EFFICACY
    bytes.fromhex(
        "59" + "0f00" + "0f00" + "ffeeddccbbaa"
    ),  # SENSOR_SETTING_SET + DEVICE_GLOBAL_TRADE_ITEM_NUMBER
    bytes.fromhex("59" + "0700" + "0700" + "64"),  # SENSOR_SETTING_SET + CHROMATICITY_TOLERANCE
    bytes.fromhex("59" + "5e00" + "5e00" + "9227"),  # SENSOR_SETTING_SET + PRESENT_PLANCKIAN_DISTANCE
    bytes.fromhex(
        "59" + "5100" + "5100" + "b80b"
    ),  # SENSOR_SETTING_SET + PRESENT_CORRELATED_COLOR_TEMPERATURE
    bytes.fromhex("59" + "0a00" + "0a00" + "04f0"),  # SENSOR_SETTING_SET + DEVICE_APPEARANCE
    bytes.fromhex("59" + "0b00" + "0b00" + "2a00"),  # SENSOR_SETTING_SET + DEVICE_COUNTRY_OF_ORIGIN
    bytes.fromhex("59" + "0c00" + "0c00" + "de4600"),  # SENSOR_SETTING_SET + DEVICE_DATE_OF_MANUFACTURE
    bytes.fromhex(
        "59" + "5000" + "5000" + "ee00cdab"
    ),  # SENSOR_SETTING_SET + PRESENT_CIE1931_CHROMATICITY_COORDINATES
    bytes.fromhex("59" + "0800" + "0800" + "64"),  # SENSOR_SETTING_SET + COLOR_RENDERING_INDEX_R9
    bytes.fromhex("59" + "0800" + "0800" + "9c"),  # SENSOR_SETTING_SET + COLOR_RENDERING_INDEX_R9
    bytes.fromhex("59" + "6100" + "6100" + "88aa00bbbb"),
    # SENSOR_SETTING_SET + RELATIVE_DEVICE_RUNTIME_IN_A_GENERIC_LEVEL_RANGE
    bytes.fromhex("59" + "6200" + "6200" + "881a27001a2700"),
    # SENSOR_SETTING_SET + RELATIVE_EXPOSURE_TIME_IN_AN_ILLUMINANCE_RANGE
    bytes.fromhex("59" + "6400" + "6400" + "88ffffe620"),
    # SENSOR_SETTING_SET + RELATIVE_RUNTIME_IN_A_DEVICE_OPERATING_TEMPERATURE_RANGE
    bytes.fromhex("59" + "6500" + "6500" + "880000cdab"),
    # SENSOR_SETTING_SET + RELATIVE_RUNTIME_IN_AN_INPUT_CURRENT_RANGE
    bytes.fromhex("59" + "6600" + "6600" + "8820032003"),
    # SENSOR_SETTING_SET + RELATIVE_RUNTIME_IN_AN_INPUT_VOLTAGE_RANGE
    # ------------------
    # Light Lightness Server
    bytes.fromhex("824b"),  # LIGHT_LIGHTNESS_GET
    bytes.fromhex("824c" + "bbaa22"),  # LIGHT_LIGHTNESS_SET
    bytes.fromhex("824c" + "010022"),  # LIGHT_LIGHTNESS_SET
    bytes.fromhex("824c" + "000031323c"),  # LIGHT_LIGHTNESS_SET
    bytes.fromhex("824d" + "000031323c"),  # LIGHT_LIGHTNESS_SET_UNACKNOWLEDGED
    bytes.fromhex("824e" + "4400"),  # LIGHT_LIGHTNESS_STATUS
    bytes.fromhex("824e" + "000031c80f"),  # LIGHT_LIGHTNESS_STATUS
    bytes.fromhex("824f"),  # LIGHT_LIGHTNESS_LINEAR_GET
    bytes.fromhex("8250" + "bbaa01"),  # LIGHT_LIGHTNESS_LINEAR_SET
    bytes.fromhex("8250" + "010022"),  # LIGHT_LIGHTNESS_LINEAR_SET
    bytes.fromhex("8250" + "000031321b"),  # LIGHT_LIGHTNESS_LINEAR_SET
    bytes.fromhex("8251" + "ff0031323c"),  # LIGHT_LIGHTNESS_LINEAR_SET_UNACKNOWLEDGED
    bytes.fromhex("8252" + "0000ddbb4c"),  # LIGHT_LIGHTNESS_LINEAR_STATUS
    bytes.fromhex("8253"),  # LIGHT_LIGHTNESS_LAST_GET
    bytes.fromhex("8254" + "0000"),  # LIGHT_LIGHTNESS_LAST_STATUS
    bytes.fromhex("8255"),  # LIGHT_LIGHTNESS_DEFAULT_GET
    bytes.fromhex("8256" + "0000"),  # LIGHT_LIGHTNESS_DEFAULT_STATUS
    bytes.fromhex("8257"),  # LIGHT_LIGHTNESS_RANGE_GET
    bytes.fromhex("8258" + "0011118888"),  # LIGHT_LIGHTNESS_RANGE_STATUS
    # ------------------
    # Scene Server
    bytes.fromhex("8241"),  # SCENE_GET
    bytes.fromhex("8242" + "01001e"),  # SCENE_RECALL
    bytes.fromhex("8243" + "01001e"),  # SCENE_RECALL_UNACKNOWLEDGED
    bytes.fromhex("8242" + "01001ef23c"),  # SCENE_RECALL
    bytes.fromhex("8243" + "01001ef23c"),  # SCENE_RECALL_UNACKNOWLEDGED
    bytes.fromhex("5e" + "000100"),  # SCENE_STATUS
    bytes.fromhex("5e" + "0001000200f2"),  # SCENE_STATUS
    bytes.fromhex("8244"),  # SCENE_REGISTER_GET
    bytes.fromhex("8245" + "00010001000200") + 14 * bytes.fromhex("0000"),  # SCENE_REGISTER_STATUS
    bytes.fromhex("8246" + "0100"),  # SCENE_STORE
    bytes.fromhex("8247" + "0100"),  # SCENE_STORE_UNACKNOWLEDGED
    bytes.fromhex("829e" + "0100"),  # SCENE_DELETE
    bytes.fromhex("829f" + "0100"),  # SCENE_DELETE_UNACKNOWLEDGED
    # ------------------
    # Generic OnOff Server
    bytes.fromhex("8201"),  # GENERIC_ONOFF_GET
    bytes.fromhex("8202" + "0122"),  # GENERIC_ONOFF_SET
    bytes.fromhex("8202" + "0122"),  # GENERIC_ONOFF_SET
    bytes.fromhex("8202" + "0031323c"),  # GENERIC_ONOFF_SET
    bytes.fromhex("8202" + "0031f23c"),  # GENERIC_ONOFF_SET
    bytes.fromhex("8204" + "00"),  # GENERIC_ONOFF_STATUS
    bytes.fromhex("8204" + "00014a"),  # GENERIC_ONOFF_STATUS
    bytes.fromhex("8204" + "0001ff"),  # GENERIC_ONOFF_STATUS
    # ------------------
    # Generic Level Server
    bytes.fromhex("8205"),  # GENERIC_LEVEL_GET
    bytes.fromhex("8206" + "ff7f22"),  # GENERIC_LEVEL_SET
    bytes.fromhex("8206" + "008022"),  # GENERIC_LEVEL_SET
    bytes.fromhex("8206" + "010022"),  # GENERIC_LEVEL_SET
    bytes.fromhex("8206" + "000031323c"),  # GENERIC_LEVEL_SET
    bytes.fromhex("8207" + "000031323c"),  # GENERIC_LEVEL_SET_UNACKNOWLEDGED
    bytes.fromhex("8208" + "ff7f"),  # GENERIC_LEVEL_STATUS
    bytes.fromhex("8208" + "0080"),  # GENERIC_LEVEL_STATUS
    bytes.fromhex("8208" + "0000ff004a"),  # GENERIC_LEVEL_STATUS
    bytes.fromhex("8208" + "00000100ff"),  # GENERIC_LEVEL_STATUS
    bytes.fromhex("8209" + "ffffff7f22"),  # GENERIC_DELTA_SET
    bytes.fromhex("8209" + "0000008022"),  # GENERIC_DELTA_SET
    bytes.fromhex("8209" + "0100000022"),  # GENERIC_DELTA_SET
    bytes.fromhex("8209" + "0000000031323c"),  # GENERIC_DELTA_SET
    bytes.fromhex("820a" + "0000000031323c"),  # GENERIC_DELTA_SET_UNACKNOWLEDGED
    bytes.fromhex("820b" + "ff7f22"),  # GENERIC_MOVE_SET
    bytes.fromhex("820b" + "008022"),  # GENERIC_MOVE_SET
    bytes.fromhex("820b" + "010022"),  # GENERIC_MOVE_SET
    bytes.fromhex("820b" + "000031323c"),  # GENERIC_MOVE_SET
    bytes.fromhex("820c" + "000031323c"),  # GENERIC_MOVE_SET_UNACKNOWLEDGED
    # ------------------
    # Time Server
    bytes.fromhex("5d" + "8ea9282a005905490248"),  # TIME_STATUS
    # ------------------
    # Silvair Debug Server V2
    bytes.fromhex("EB3601" + "02" + "090002001020300208000420"),  # STATUS
    bytes.fromhex("EB3601" + "02" + "0b020c06001000200030"),  # STATUS
    bytes.fromhex("EB3601" + "02" + "4b000b096d7920737472696e67"),  # STATUS
    bytes.fromhex("EB3601" + "02" + "550002001020300a000800320028"),  # STATUS
    bytes.fromhex("EB3601" + "03" + "1300"),  # CLEAR
    bytes.fromhex("EB3601" + "00" + "0f00"),  # GET
    bytes.fromhex("EB3601" + "02" + "fe00ff"),  # STATUS
    # Configuration Client
    bytes.fromhex("02" + "01" + "010000"),  # CONFIG_COMPOSITION_DATA_STATUS
    bytes.fromhex("02" + "01" + "01000400"),  # CONFIG_COMPOSITION_DATA_STATUS
    bytes.fromhex("02" + "01" + "02010005000000010201000501170101"),  # CONFIG_COMPOSITION_DATA_STATUS
    bytes.fromhex(
        "02" + "01" + "050307011AFD0901383001080508480000050A00050AD0"
    ),  # CONFIG_COMPOSITION_DATA_STATUS
    bytes.fromhex("8028" + "0200"),  # CONFIG_RELAY_STATUS
    bytes.fromhex("8028" + "02FF"),  # CONFIG_RELAY_STATUS
    bytes.fromhex("8028" + "0100"),  # CONFIG_RELAY_STATUS
    bytes.fromhex("8028" + "0000"),  # CONFIG_RELAY_STATUS
    bytes.fromhex("8028" + "01FF"),  # CONFIG_RELAY_STATUS
    bytes.fromhex("8028" + "00FF"),  # CONFIG_RELAY_STATUS
    bytes.fromhex("8027" + "0100"),  # CONFIG_RELAY_SET
    bytes.fromhex("8027" + "0000"),  # CONFIG_RELAY_SET
    bytes.fromhex("8027" + "01FF"),  # CONFIG_RELAY_SET
    bytes.fromhex("8027" + "00FF"),  # CONFIG_RELAY_SET
    # ------------------
    # Generic User Property Server
    bytes.fromhex("822E"),  # Generic User Properties Get
    bytes.fromhex(
        "4B"
        + "7300"
        + "8800"
        + "8900"
        + "8C00"
        + "8D00"
        + "8E00"
        + "8F00"
        + "9000"
        + "9100"
        + "9200"
        + "9300"
        + "9400"
        + "9500"
        + "9600"
    ),  # Generic User Properties Status
    bytes.fromhex("822F" + "6F00"),  # Generic User Property Get + Total Light Exposure Time
    bytes.fromhex("4C" + "8C00" + "03"),  # Generic User Property Set + Light Distribution + Type_III
    bytes.fromhex(
        "4D" + "8800" + "FEFFFF"
    ),  # Generic User Property Set Unacknowledged + External Supply Voltage + 262143.97 V
    bytes.fromhex(
        "4E" + "8900" + "01" + "FEFF"
    ),  # Generic User Property Status + External Supply Voltage Frequency + Read Only + 65533 Hz
    # Generic Admin Property Server
    bytes.fromhex("822C"),  # Generic Admin Properties Get
    bytes.fromhex(
        "47"
        + "9700"
        + "9800"
        + "9900"
        + "9A00"
        + "9C00"
        + "9D00"
        + "9E00"
        + "9F00"
        + "A000"
        + "A400"
        + "A500"
        + "AB00"
        + "AC00"
        + "AD00"
    ),  # Generic Admin Properties Status
    bytes.fromhex("822D" + "7000"),  # Generic Admin Property Get + Total Luminous Energy
    bytes.fromhex(
        "48" + "9400" + "00" + "7fff"
    ),  # Generic Admin Property Set + Light Source Temperature  + Not a Generic User Property + 16383.5 °C
    bytes.fromhex(
        "49" + "A400" + "02" + "FFFFFD"
    ),  # Generic Admin Property Set Unacknowledged + Nominal Light Output + Write Only + 16777213 lm
    bytes.fromhex(
        "4A" + "7300" + "02" + "64"
    ),  # Generic Admin Property Status + Write Only + Power Factor + 100
    # ------------------
    # Generic Manufacturer Property Server
    bytes.fromhex("822A"),  # Generic Manufacturer Properties Get
    bytes.fromhex(
        "43" + "AE00" + "B400" + "B500" + "B600" + "B700"
    ),  # Generic Manufacturer Properties Status
    bytes.fromhex("822B" + "7200"),  # Generic Manufacturer Property Get + Precise Total Device Energy Use
    bytes.fromhex(
        "44" + "8D00" + "00"
    ),  # Generic Manufacturer Property Set + Light Source Current + Not a Generic User Property
    bytes.fromhex(
        "45" + "9900" + "01"
    ),  # Generic Manufacturer Property Set Unacknowledged + Luminaire Color + Generic User Property and can be read
    bytes.fromhex(
        "46" + "A000" + "03" + "FEFF00"
    ),  # Generic Manufacturer Property Status + Luminaire Time Of Manufacture + Read Write + 12-12-12 2012 12:12
    bytes.fromhex(
        "46" + "8300" + "01" + "FDFFFFFF"
    ),  # Generic Manufacturer Property Status + Apparent Energy + User Access + 4294967.293 kVAh
    bytes.fromhex(
        "46" + "8400" + "01" + "FDFFFF"
    ),  # Generic Manufacturer Property Status + Apparent Power + User Access + 1677721.3 VA
    bytes.fromhex(
        "46" + "8000" + "01" + "FDFFFFFF"
    ),  # Generic Manufacturer Property Status + Active Energy Loadside + User Access + 4294967.293 kWh
    bytes.fromhex(
        "46" + "B300" + "03" + "05"
    ),  # Generic Manufacturer Property Status + Light Source Type + Read Write + Light emitting diode (LED)
    # ------------------
    # Generic Client Property Server
    bytes.fromhex("4F"),  # Generic Client Properties Get
    bytes.fromhex("50" + "7300"),  # Generic Client Properties Status
    # ------------------
    # Silvair Light Extended Controller
    bytes.fromhex("f63601" + "00" + "71ff"),  # PROPERTY_GET
    bytes.fromhex("f63601" + "01" + "71ff01"),  # PROPERTY_SET
    bytes.fromhex("f63601" + "03" + "72ff90ca04"),  # PROPERTY_STATUS
    # ------------------
    # Silvair EL Server
    bytes.fromhex("EA3601" + "00"),  # EL_INHIBIT_ENTER
    bytes.fromhex("EA3601" + "01"),  # EL_INHIBIT_ENTER_UNACKNOWLEDGED
    bytes.fromhex("EA3601" + "02"),  # EL_INHIBIT_EXIT
    bytes.fromhex("EA3601" + "03"),  # EL_INHIBIT_EXIT_UNACKNOWLEDGED
    bytes.fromhex("EA3601" + "04"),  # EL_STATE_GET
    bytes.fromhex("EA3601" + "05" + "00"),  # EL_STATE_STATUS
    bytes.fromhex("EA3601" + "06" + "80FF"),  # EL_PROPERTY_GET
    bytes.fromhex("EA3601" + "07" + "80FF3412"),  # EL_PROPERTY_SET
    bytes.fromhex("EA3601" + "08" + "80FF3412"),  # EL_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("EA3601" + "09" + "80FF3412"),  # EL_PROPERTY_STATUS
    bytes.fromhex("EA3601" + "0A"),  # EL_LAMP_OPERATION_TIME_GET
    bytes.fromhex("EA3601" + "0B"),  # EL_LAMP_OPERATION_TIME_CLEAR
    bytes.fromhex("EA3601" + "0C"),  # EL_LAMP_OPERATION_TIME_CLEAR_UNACKNOWLEDGED
    bytes.fromhex("EA3601" + "0D" + "67452301EFCDAB89"),  # EL_LAMP_OPERATION_TIME_STATUS
    bytes.fromhex("EA3601" + "0E"),  # EL_REST_ENTER
    bytes.fromhex("EA3601" + "0F"),  # EL_REST_ENTER_UNACKNOWLEDGED
    bytes.fromhex("EA3601" + "10"),  # EL_REST_EXIT
    bytes.fromhex("EA3601" + "11"),  # EL_REST_EXIT_UNACKNOWLEDGED
    # ------------------
    # Silvair ELT Server
    bytes.fromhex("E93601" + "00"),  # ELT_FUNCTIONAL_TEST_GET
    bytes.fromhex("E93601" + "01"),  # ELT_FUNCTIONAL_TEST_START
    bytes.fromhex("E93601" + "02"),  # ELT_FUNCTIONAL_TEST_STOP
    bytes.fromhex("E93601" + "03" + "0000000000401F010000"),  # ELT_FUNCTIONAL_TEST_STATUS
    bytes.fromhex("E93601" + "03" + "000000000040FF000701"),  # ELT_FUNCTIONAL_TEST_STATUS
    bytes.fromhex("E93601" + "03" + "2E48D52F004824010202"),  # ELT_FUNCTIONAL_TEST_STATUS
    bytes.fromhex("E93601" + "03" + "0000000000782401030745000000"),  # ELT_FUNCTIONAL_TEST_STATUS
    bytes.fromhex("E93601" + "04"),  # ELT_DURATION_TEST_GET
    bytes.fromhex("E93601" + "05"),  # ELT_DURATION_TEST_START
    bytes.fromhex("E93601" + "06"),  # ELT_DURATION_TEST_STOP
    bytes.fromhex("E93601" + "07" + "0000000000401F0100000000"),  # ELT_DURATION_TEST_STATUS
    bytes.fromhex("E93601" + "07" + "000000000040FF000701FFFF"),  # ELT_DURATION_TEST_STATUS
    bytes.fromhex("E93601" + "07" + "2E48D52F0048240102020000"),  # ELT_DURATION_TEST_STATUS
    bytes.fromhex("E93601" + "07" + "0000000000782401030F100E45000000"),  # ELT_DURATION_TEST_STATUS
    bytes.fromhex("E93601" + "08" + "84FF"),  # ELT_PROPERTY_GET
    bytes.fromhex("E93601" + "08" + "85FF"),  # ELT_PROPERTY_GET
    bytes.fromhex("E93601" + "08" + "86FF"),  # ELT_PROPERTY_GET
    bytes.fromhex("E93601" + "08" + "87FF"),  # ELT_PROPERTY_GET
    bytes.fromhex("E93601" + "08" + "88FF"),  # ELT_PROPERTY_GET
    bytes.fromhex("E93601" + "08" + "89FF"),  # ELT_PROPERTY_GET
    bytes.fromhex("E93601" + "08" + "8AFF"),  # ELT_PROPERTY_GET
    bytes.fromhex("E93601" + "08" + "8BFF"),  # ELT_PROPERTY_GET
    bytes.fromhex("E93601" + "09" + "84FF04030201"),  # ELT_PROPERTY_SET
    bytes.fromhex("E93601" + "09" + "85FF04030201"),  # ELT_PROPERTY_SET
    bytes.fromhex("E93601" + "09" + "86FF04030201"),  # ELT_PROPERTY_SET
    bytes.fromhex("E93601" + "09" + "87FF04030201"),  # ELT_PROPERTY_SET
    bytes.fromhex("E93601" + "09" + "88FF04030201"),  # ELT_PROPERTY_SET
    bytes.fromhex("E93601" + "09" + "89FF04030201"),  # ELT_PROPERTY_SET
    bytes.fromhex("E93601" + "09" + "8AFF04030201"),  # ELT_PROPERTY_SET
    bytes.fromhex("E93601" + "09" + "8BFF04030201"),  # ELT_PROPERTY_SET
    bytes.fromhex("E93601" + "0A" + "84FF04030201"),  # ELT_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("E93601" + "0A" + "85FF04030201"),  # ELT_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("E93601" + "0A" + "86FF04030201"),  # ELT_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("E93601" + "0A" + "87FF04030201"),  # ELT_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("E93601" + "0A" + "88FF04030201"),  # ELT_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("E93601" + "0A" + "89FF04030201"),  # ELT_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("E93601" + "0A" + "8AFF04030201"),  # ELT_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("E93601" + "0A" + "8BFF04030201"),  # ELT_PROPERTY_SET_UNACKNOWLEDGED
    bytes.fromhex("E93601" + "0B" + "84FF04030201"),  # ELT_PROPERTY_STATUS
    bytes.fromhex("E93601" + "0B" + "85FF04030201"),  # ELT_PROPERTY_STATUS
    bytes.fromhex("E93601" + "0B" + "86FF04030201"),  # ELT_PROPERTY_STATUS
    bytes.fromhex("E93601" + "0B" + "87FF04030201"),  # ELT_PROPERTY_STATUS
    bytes.fromhex("E93601" + "0B" + "88FF04030201"),  # ELT_PROPERTY_STATUS
    bytes.fromhex("E93601" + "0B" + "89FF04030201"),  # ELT_PROPERTY_STATUS
    bytes.fromhex("E93601" + "0B" + "8AFF04030201"),  # ELT_PROPERTY_STATUS
    bytes.fromhex("E93601" + "0B" + "8BFF04030201"),  # ELT_PROPERTY_STATUS
]


@pytest.fixture(scope="session")
def capnproto():
    import capnp

    from bluetooth_mesh.messages.capnproto import generate

    with NamedTemporaryFile("w", suffix=".capnp") as f:
        generate(0xD988DA1AAFBE9E47, f)
        f.flush()
        return capnp.load(f.name)


@pytest.mark.skipif(not importlib.util.find_spec("capnp"), reason="requires Python3.7")
@pytest.mark.parametrize("encoded", [pytest.param(i, id=i.hex()) for i in valid])
def test_parse_capnproto(encoded, capnproto):
    logging.info("MESH[%i] %s", len(encoded), encoded.hex())

    decoded = AccessMessage.parse(encoded)
    logging.info("CONSTRUCT %r", decoded)

    params = to_camelcase_dict(decoded)
    logging.info("CAPNP INPUT[%i] %s", len(json.dumps(params)), json.dumps(params))

    message = capnproto.AccessMessage.new_message(**params)
    logging.info("CAPNP %r", message)

    packed = message.to_bytes_packed()
    logging.info("PACKED[%i] %s", len(packed), packed.hex())

    unpacked = capnproto.AccessMessage.from_bytes_packed(packed)
    logging.info("UNPACKED %r", unpacked)

    params = to_snakecase_dict(unpacked.to_dict())
    logging.info("CONSTRUCT INPUT %s", params)

    recoded = AccessMessage.build(params)
    logging.info("RECODED[%i] %s", len(recoded), recoded.hex())

    assert recoded == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2024  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import Construct, Embedded, GreedyRange, Int8ul, Int16ul, Struct, this

from bluetooth_mesh.messages.properties import PropertyDict, PropertyID, PropertyMixin
from bluetooth_mesh.messages.sensor import SensorPropertyId
from bluetooth_mesh.messages.util import EnumAdapter, EnumSwitch as Switch, Opcode, SwitchStruct


class GenericPropertyOpcode(IntEnum):
    GENERIC_USER_PROPERTIES_GET = 0x822E
    GENERIC_USER_PROPERTIES_STATUS = 0x4B
    GENERIC_USER_PROPERTY_GET = 0x822F
    GENERIC_USER_PROPERTY_SET = 0x4C
    GENERIC_USER_PROPERTY_SET_UNACKNOWLEDGED = 0x4D
    GENERIC_USER_PROPERTY_STATUS = 0x4E
    GENERIC_ADMIN_PROPERTIES_GET = 0x822C
    GENERIC_ADMIN_PROPERTIES_STATUS = 0x47
    GENERIC_ADMIN_PROPERTY_GET = 0x822D
    GENERIC_ADMIN_PROPERTY_SET = 0x48
    GENERIC_ADMIN_PROPERTY_SET_UNACKNOWLEDGED = 0x49
    GENERIC_ADMIN_PROPERTY_STATUS = 0x4A
    GENERIC_MANUFACTURER_PROPERTIES_GET = 0x822A
    GENERIC_MANUFACTURER_PROPERTIES_STATUS = 0x43
    GENERIC_MANUFACTURER_PROPERTY_GET = 0x822B
    GENERIC_MANUFACTURER_PROPERTY_SET = 0x44
    GENERIC_MANUFACTURER_PROPERTY_SET_UNACKNOWLEDGED = 0x45
    GENERIC_MANUFACTURER_PROPERTY_STATUS = 0x46
    GENERIC_CLIENT_PROPERTIES_GET = 0x4F
    GENERIC_CLIENT_PROPERTIES_STATUS = 0x50


class UserAccess(IntEnum):
    READ_ONLY = 0x01
    WRITE_ONLY = 0x02
    READ_WRITE = 0x03


class AdminUserAccess(IntEnum):
    NOT_USER_PROPERTY = 0x00
    READ_ONLY = 0x01
    WRITE_ONLY = 0x02
    READ_WRITE = 0x03


class ManufacturerUserAccess(IntEnum):
    NOT_USER_PROPERTY = 0x00
    READ_ONLY = 0x01


GenericPropertiesGet = Struct()

GenericPropertiesStatus = Struct(
    "property_ids" / GreedyRange(SensorPropertyId),
)

GenericPropertyGet = Struct(
    "property_id" / SensorPropertyId,
)


class _GenericUserPropertySet(PropertyMixin, Construct):
    ENUM = PropertyID
    DICT = PropertyDict

    subcon = Struct(
        "property_id" / EnumAdapter(Int16ul, PropertyID),
        Switch(this.id, PropertyDict),
    )

    def _parse(self, stream, context, path):
        msg = GenericPropertyGet._parse(stream, context, path)
        return self._parse_property(msg, stream, context, path)

    def _build(self, obj, stream, context, path):
        GenericPropertyGet._build(obj, stream, context, path)
        return self._build_property(obj, stream, context, path)


class _GenericPropertyStatus(PropertyMixin, Construct):
    ENUM = PropertyID
    DICT = PropertyDict

    subcon = Struct(
        "property_id" / EnumAdapter(Int16ul, PropertyID),
        "access" / EnumAdapter(Int8ul, UserAccess),
        Switch(this.id, PropertyDict),
    )

    def _parse(self, stream, context, path):
        msg = Struct(Embedded(GenericPropertyGet), "access" / Int8ul)._parse(stream, context, path)
        return self._parse_property(msg, stream, context, path)

    def _build(self, obj, stream, context, path):
        Struct(Embedded(GenericPropertyGet), "access" / Int8ul)._build(obj, stream, context, path)
        return self._build_property(obj, stream, context, path)


GenericUserPropertySet = _GenericUserPropertySet()

GenericAdminPropertySet = _GenericPropertyStatus()

GenericManufacturerPropertySet = Struct(
    "property_id" / EnumAdapter(Int16ul, PropertyID),
    "access" / EnumAdapter(Int8ul, ManufacturerUserAccess),
)

GenericPropertyStatus = _GenericPropertyStatus()

GenericPropertyMessage = SwitchStruct(
    "opcode" / Opcode(GenericPropertyOpcode),
    "params"
    / Switch(
        this.opcode,
        {
            GenericPropertyOpcode.GENERIC_USER_PROPERTIES_GET: GenericPropertiesGet,
            GenericPropertyOpcode.GENERIC_USER_PROPERTIES_STATUS: GenericPropertiesStatus,
            GenericPropertyOpcode.GENERIC_USER_PROPERTY_GET: GenericPropertyGet,
            GenericPropertyOpcode.GENERIC_USER_PROPERTY_SET: GenericUserPropertySet,
            GenericPropertyOpcode.GENERIC_USER_PROPERTY_SET_UNACKNOWLEDGED: GenericUserPropertySet,
            GenericPropertyOpcode.GENERIC_USER_PROPERTY_STATUS: GenericPropertyStatus,
            GenericPropertyOpcode.GENERIC_ADMIN_PROPERTIES_GET: GenericPropertiesGet,
            GenericPropertyOpcode.GENERIC_ADMIN_PROPERTIES_STATUS: GenericPropertiesStatus,
            GenericPropertyOpcode.GENERIC_ADMIN_PROPERTY_GET: GenericPropertyGet,
            GenericPropertyOpcode.GENERIC_ADMIN_PROPERTY_SET: GenericAdminPropertySet,
            GenericPropertyOpcode.GENERIC_ADMIN_PROPERTY_SET_UNACKNOWLEDGED: GenericAdminPropertySet,
            GenericPropertyOpcode.GENERIC_ADMIN_PROPERTY_STATUS: GenericPropertyStatus,
            GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTIES_GET: GenericPropertiesGet,
            GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTIES_STATUS: GenericPropertiesStatus,
            GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_GET: GenericPropertyGet,
            GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_SET: GenericManufacturerPropertySet,
            GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_SET_UNACKNOWLEDGED: GenericManufacturerPropertySet,
            GenericPropertyOpcode.GENERIC_MANUFACTURER_PROPERTY_STATUS: GenericPropertyStatus,
            GenericPropertyOpcode.GENERIC_CLIENT_PROPERTIES_GET: GenericPropertiesGet,
            GenericPropertyOpcode.GENERIC_CLIENT_PROPERTIES_STATUS: GenericPropertiesStatus,
        },
    ),
)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from datetime import timedelta

import pytest

from bluetooth_mesh.messages.silvair.light_extended_controller import (
    LightExtendedControllerMessage,
    LightExtendedControllerOpcode,
    LightExtendedControllerProperty,
    LightExtendedControllerSubOpcode,
)

valid = [
    # fmt: off
    pytest.param(
        bytes.fromhex("f636010071ff"),
        LightExtendedControllerOpcode.SILVAIR_LEC,
        dict(
            subopcode=LightExtendedControllerSubOpcode.PROPERTY_GET,
            payload=dict(
                property_id=LightExtendedControllerProperty.AUTO_RESUME_MODE,
            )
        ),
        id="LIGHT_LEC_PROPERTY_GET[AUTO_RESUME_MODE]"
    ),
    pytest.param(
        bytes.fromhex("f636010171ff01"),
        LightExtendedControllerOpcode.SILVAIR_LEC,
        dict(
            subopcode=LightExtendedControllerSubOpcode.PROPERTY_SET,
            payload=dict(
                property_id=LightExtendedControllerProperty.AUTO_RESUME_MODE,
                auto_resume_mode=True,
            )
        ),
        id="LIGHT_LEC_PROPERTY_SET[AUTO_RESUME_MODE]"
    ),
    pytest.param(
        bytes.fromhex("f636010272ff10A400"),
        LightExtendedControllerOpcode.SILVAIR_LEC,
        dict(
            subopcode=LightExtendedControllerSubOpcode.PROPERTY_SET_UNACKNOWLEDGED,
            payload=dict(
                property_id=LightExtendedControllerProperty.AUTO_RESUME_TIMER,
                auto_resume_timer=dict(seconds=42),
            )
        ),
        id="LIGHT_LEC_PROPERTY_SET_UNACKNOWLEDGED[AUTO_RESUME_TIMER]"
    ),
    pytest.param(
        bytes.fromhex("f636010372ff90ca04"),
        LightExtendedControllerOpcode.SILVAIR_LEC,
        dict(
            subopcode=LightExtendedControllerSubOpcode.PROPERTY_STATUS,
            payload=dict(
                property_id=LightExtendedControllerProperty.AUTO_RESUME_TIMER,
                auto_resume_timer=dict(seconds=314),
            )
        ),
        id="LIGHT_LEC_PROPERTY_STATUS[AUTO_RESUME_TIMER]"
    ),
    # fmt: on
]


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_parse_valid(encoded, opcode, data):
    LightExtendedControllerMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_build_valid(encoded, opcode, data):
    assert LightExtendedControllerMessage.build(dict(opcode=opcode, params=data)) == encoded



import pytest

from bluetooth_mesh.messages.silvair.emergency_lighting import (
    ElState,
    EmergencyLightingParams,
    EmergencyLightingProperty,
    EmergencyLightingSubOpcode,
)

valid = [
    pytest.param(
        bytes.fromhex("00"),
        EmergencyLightingSubOpcode.EL_INHIBIT_ENTER,
        {},
    ),
    pytest.param(
        bytes.fromhex("01"),
        EmergencyLightingSubOpcode.EL_INHIBIT_ENTER_UNACKNOWLEDGED,
        {},
    ),
    pytest.param(
        bytes.fromhex("02"),
        EmergencyLightingSubOpcode.EL_INHIBIT_EXIT,
        {},
    ),
    pytest.param(
        bytes.fromhex("03"),
        EmergencyLightingSubOpcode.EL_INHIBIT_EXIT_UNACKNOWLEDGED,
        {},
    ),
    pytest.param(
        bytes.fromhex("04"),
        EmergencyLightingSubOpcode.EL_STATE_GET,
        {},
    ),
    pytest.param(
        bytes.fromhex("0500"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.TRANSITION_FROM_INHIBIT_TO_NORMAL,
        ),
    ),
    pytest.param(
        bytes.fromhex("0501"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.TRANSITION_FROM_FUNCTIONAL_TEST_TO_NORMAL,
        ),
    ),
    pytest.param(
        bytes.fromhex("0502"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.TRANSITION_FROM_DURATION_TEST_TO_NORMAL,
        ),
    ),
    pytest.param(
        bytes.fromhex("0503"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.NORMAL,
        ),
    ),
    pytest.param(
        bytes.fromhex("0504"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.TRANSITION_FROM_REST_TO_EMERGENCY,
        ),
    ),
    pytest.param(
        bytes.fromhex("0505"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.EMERGENCY,
        ),
    ),
    pytest.param(
        bytes.fromhex("0506"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.EXTENDED_EMERGENCY,
        ),
    ),
    pytest.param(
        bytes.fromhex("0507"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.TRANSITION_FROM_EMERGENCY_TO_REST,
        ),
    ),
    pytest.param(
        bytes.fromhex("0508"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.REST,
        ),
    ),
    pytest.param(
        bytes.fromhex("0509"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.TRANSITION_FROM_NORMAL_TO_INHIBIT,
        ),
    ),
    pytest.param(
        bytes.fromhex("050A"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.INHIBIT,
        ),
    ),
    pytest.param(
        bytes.fromhex("050B"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.TRANSITION_FROM_NORMAL_TO_DURATION_TEST,
        ),
    ),
    pytest.param(
        bytes.fromhex("050C"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.DURATION_TEST_IN_PROGRESS,
        ),
    ),
    pytest.param(
        bytes.fromhex("050D"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.TRANSITION_FROM_NORMAL_TO_FUNCTIONAL_TEST,
        ),
    ),
    pytest.param(
        bytes.fromhex("050E"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.FUNCTIONAL_TEST_IN_PROGRESS,
        ),
    ),
    pytest.param(
        bytes.fromhex("050F"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.BATTERY_DISCHARGED,
        ),
    ),
    pytest.param(
        bytes.fromhex("0510"),
        EmergencyLightingSubOpcode.EL_STATE_STATUS,
        dict(
            el_state=ElState.UNKNOWN,
        ),
    ),
    pytest.param(
        bytes.fromhex("0680FF"),
        EmergencyLightingSubOpcode.EL_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS,
        ),
    ),
    pytest.param(
        bytes.fromhex("0681FF"),
        EmergencyLightingSubOpcode.EL_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MIN,
        ),
    ),
    pytest.param(
        bytes.fromhex("0682FF"),
        EmergencyLightingSubOpcode.EL_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MAX,
        ),
    ),
    pytest.param(
        bytes.fromhex("0683FF"),
        EmergencyLightingSubOpcode.EL_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingProperty.EL_PROLONG_TIME,
        ),
    ),
    pytest.param(
        bytes.fromhex("0780FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS,
            el_lightness=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0781FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MIN,
            el_lightness_range_min=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0782FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MAX,
            el_lightness_range_max=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0783FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingProperty.EL_PROLONG_TIME,
            el_prolong_time=dict(seconds=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0880FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS,
            el_lightness=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0881FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MIN,
            el_lightness_range_min=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0882FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MAX,
            el_lightness_range_max=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0883FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingProperty.EL_PROLONG_TIME,
            el_prolong_time=dict(seconds=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0980FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS,
            el_lightness=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0981FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MIN,
            el_lightness_range_min=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0982FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MAX,
            el_lightness_range_max=dict(perceived_lightness=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0983FF3412"),
        EmergencyLightingSubOpcode.EL_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingProperty.EL_PROLONG_TIME,
            el_prolong_time=dict(seconds=0x1234),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A"),
        EmergencyLightingSubOpcode.EL_LAMP_OPERATION_TIME_GET,
        {},
    ),
    pytest.param(
        bytes.fromhex("0B"),
        EmergencyLightingSubOpcode.EL_LAMP_OPERATION_TIME_CLEAR,
        {},
    ),
    pytest.param(
        bytes.fromhex("0C"),
        EmergencyLightingSubOpcode.EL_LAMP_OPERATION_TIME_CLEAR_UNACKNOWLEDGED,
        {},
    ),
    pytest.param(
        bytes.fromhex("0D67452301EFCDAB89"),
        EmergencyLightingSubOpcode.EL_LAMP_OPERATION_TIME_STATUS,
        dict(
            el_lamp_total_operation_time=0x01234567,
            el_lamp_emergency_time=0x89ABCDEF,
        ),
    ),
    pytest.param(
        bytes.fromhex("0E"),
        EmergencyLightingSubOpcode.EL_REST_ENTER,
        {},
    ),
    pytest.param(
        bytes.fromhex("0F"),
        EmergencyLightingSubOpcode.EL_REST_ENTER_UNACKNOWLEDGED,
        {},
    ),
    pytest.param(
        bytes.fromhex("10"),
        EmergencyLightingSubOpcode.EL_REST_EXIT,
        {},
    ),
    pytest.param(
        bytes.fromhex("11"),
        EmergencyLightingSubOpcode.EL_REST_EXIT_UNACKNOWLEDGED,
        {},
    ),
]


@pytest.mark.parametrize("encoded, subopcode, payload", valid)
def test_parse(encoded, subopcode, payload):
    decoded = EmergencyLightingParams.parse(encoded)
    assert decoded.subopcode == subopcode
    assert decoded.payload == payload


@pytest.mark.parametrize("encoded, subopcode, payload", valid)
def test_build(encoded, subopcode, payload):
    assert EmergencyLightingParams.build(dict(subopcode=subopcode, payload=payload)) == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import Construct, Flag, Int8ul, Int16ul, Struct, this

from bluetooth_mesh.messages.properties import PropertyMixin, TimeMiliseconds24
from bluetooth_mesh.messages.util import EnumAdapter, EnumSwitch as Switch, Opcode, SwitchStruct


class LightExtendedControllerOpcode(IntEnum):
    SILVAIR_LEC = 0xF63601


class LightExtendedControllerSubOpcode(IntEnum):
    PROPERTY_GET = 0x00
    PROPERTY_SET = 0x01
    PROPERTY_SET_UNACKNOWLEDGED = 0x02
    PROPERTY_STATUS = 0x03
    BULK_LC_PROPERTY_SET = 0x04
    BULK_LC_PROPERTY_STATUS = 0x05
    BULK_LEC_PROPERTY_SET = 0x06
    BULK_LEC_PROPERTY_STATUS = 0x06
    SYNC_INTEGRAL_GET = 0x08
    SYNC_INTEGRAL_STATUS = 0x09


class LightExtendedControllerProperty(IntEnum):
    AUTO_RESUME_MODE = 0xFF71
    AUTO_RESUME_TIMER = 0xFF72


LightExtendedControllerPropertyDict = {
    LightExtendedControllerProperty.AUTO_RESUME_MODE: Flag,
    LightExtendedControllerProperty.AUTO_RESUME_TIMER: TimeMiliseconds24,
}


class _LightExtendedControllerProperty(PropertyMixin, Construct):
    ENUM = LightExtendedControllerProperty
    DICT = LightExtendedControllerPropertyDict

    subcon = Struct(
        "property_id" / EnumAdapter(Int16ul, LightExtendedControllerProperty),
        Switch(this.id, LightExtendedControllerPropertyDict),
    )

    def _parse(self, stream, context, path):
        msg = LightExtendedControllerPropertyGet._parse(stream, context, path)
        return self._parse_property(msg, stream, context, path)

    def _build(self, obj, stream, context, path):
        LightExtendedControllerPropertyGet._build(obj, stream, context, path)
        return self._build_property(obj, stream, context, path)


# fmt: off
LightExtendedControllerPropertyGet = Struct(
    "property_id" / EnumAdapter(Int16ul, LightExtendedControllerProperty),
)

LightExtendedControllerPropertySet = _LightExtendedControllerProperty()

LightExtendedControllerPropertyStatus = _LightExtendedControllerProperty()

LightExtendedControllerSyncIntegralStatus = Struct(
    "sync_integral" / Int16ul,
)

LightExtendedControllerParams = SwitchStruct(
    "subopcode" / EnumAdapter(Int8ul, LightExtendedControllerSubOpcode),
    "payload" / Switch(
        this.subopcode,
        {
            LightExtendedControllerSubOpcode.PROPERTY_GET: LightExtendedControllerPropertyGet,
            LightExtendedControllerSubOpcode.PROPERTY_SET: LightExtendedControllerPropertySet,
            LightExtendedControllerSubOpcode.PROPERTY_SET_UNACKNOWLEDGED: LightExtendedControllerPropertySet,
            LightExtendedControllerSubOpcode.PROPERTY_STATUS: LightExtendedControllerPropertyStatus,
            LightExtendedControllerSubOpcode.SYNC_INTEGRAL_STATUS: LightExtendedControllerSyncIntegralStatus,
        }
    )
)

LightExtendedControllerMessage = SwitchStruct(
    "opcode" / Opcode(LightExtendedControllerOpcode),
    "params" / Switch(
        this.opcode,
        {
            LightExtendedControllerOpcode.SILVAIR_LEC: LightExtendedControllerParams,
        }
    )
)
# fmt: on



from datetime import datetime, timedelta, timezone

import pytest

from bluetooth_mesh.messages.silvair.emergency_lighting_test import (
    EmergencyLightingTestParams,
    EmergencyLightingTestProperty,
    EmergencyLightingTestSubOpcode,
    TestExecutionStatus,
)

valid = [
    pytest.param(
        bytes.fromhex("00"),
        EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_GET,
        {},
    ),
    pytest.param(
        bytes.fromhex("01"),
        EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_START,
        {},
    ),
    pytest.param(
        bytes.fromhex("02"),
        EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_STOP,
        {},
    ),
    pytest.param(
        bytes.fromhex(
            "03"  # Subopcode
            "0000000000"  # ELT Test TAI Timestamp: TAI Seconds
            "40"  # ELT Test TAI Timestamp: Time Zone Offset
            "1F01"  # ELT Test TAI Timestamp: TAI-UTC Delta
            "00"  # ELT Test Execution Status
            "00"  # ELT Test Execution Result
        ),
        EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    1999,
                    12,
                    31,
                    23,
                    59,
                    28,
                    tzinfo=timezone(timedelta(hours=0)),
                ),
                tai_utc_delta=timedelta(seconds=32),
            ),
            execution_result=dict(
                lamp_fault=False,
                battery_fault=False,
                circuit_fault=False,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.FINISHED,
        ),
    ),
    pytest.param(
        bytes.fromhex(
            "03"  # Subopcode
            "0000000000"  # ELT Test TAI Timestamp: TAI Seconds
            "40"  # ELT Test TAI Timestamp: Time Zone Offset
            "FF00"  # ELT Test TAI Timestamp: TAI-UTC Delta
            "07"  # ELT Test Execution Status
            "01"  # ELT Test Execution Result
        ),
        EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    2000,
                    1,
                    1,
                    0,
                    0,
                    0,
                    tzinfo=timezone(timedelta(hours=0)),
                ),
                tai_utc_delta=timedelta(seconds=0),
            ),
            execution_result=dict(
                lamp_fault=True,
                battery_fault=False,
                circuit_fault=False,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.UNKNOWN,
        ),
    ),
    pytest.param(
        bytes.fromhex(
            "03"  # Subopcode
            "2E48D52F00"  # ELT Test TAI Timestamp: TAI Seconds
            "48"  # ELT Test TAI Timestamp: Time Zone Offset
            "2401"  # ELT Test TAI Timestamp: TAI-UTC Delta
            "02"  # ELT Test Execution Status
            "02"  # ELT Test Execution Result
        ),
        EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    2025,
                    6,
                    6,
                    8,
                    32,
                    41,
                    tzinfo=timezone(timedelta(hours=2)),
                ),
                tai_utc_delta=timedelta(seconds=37),
            ),
            execution_result=dict(
                lamp_fault=False,
                battery_fault=True,
                circuit_fault=False,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.IN_PROGRESS,
        ),
    ),
    pytest.param(
        bytes.fromhex(
            "03"  # Subopcode
            "0000000000"  # ELT Test TAI Timestamp: TAI Seconds
            "78"  # ELT Test TAI Timestamp: Time Zone Offset
            "2401"  # ELT Test TAI Timestamp: TAI-UTC Delta
            "03"  # ELT Test Execution Status
            "07"  # ELT Test Execution Result
            "45000000"  # ELT Test Relative Timestamp (optional)
        ),
        EmergencyLightingTestSubOpcode.ELT_FUNCTIONAL_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    2000,
                    1,
                    1,
                    13,
                    59,
                    23,
                    tzinfo=timezone(timedelta(hours=14)),
                ),
                tai_utc_delta=timedelta(seconds=37),
            ),
            execution_result=dict(
                lamp_fault=True,
                battery_fault=True,
                circuit_fault=True,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.POSTPONED,
            relative_timestamp=69,
        ),
    ),
    pytest.param(
        bytes.fromhex("04"),
        EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_GET,
        {},
    ),
    pytest.param(
        bytes.fromhex("05"),
        EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_START,
        {},
    ),
    pytest.param(
        bytes.fromhex("06"),
        EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_STOP,
        {},
    ),
    pytest.param(
        bytes.fromhex(
            "07"  # Subopcode
            "0000000000"  # ELT Test TAI Timestamp: TAI Seconds
            "40"  # ELT Test TAI Timestamp: Time Zone Offset
            "1F01"  # ELT Test TAI Timestamp: TAI-UTC Delta
            "00"  # ELT Test Execution Status
            "00"  # ELT Test Execution Result
            "0000"  # ELT Duration Result
        ),
        EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    1999,
                    12,
                    31,
                    23,
                    59,
                    28,
                    tzinfo=timezone(timedelta(hours=0)),
                ),
                tai_utc_delta=timedelta(seconds=32),
            ),
            execution_result=dict(
                lamp_fault=False,
                battery_fault=False,
                circuit_fault=False,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.FINISHED,
            duration_result=0,
        ),
    ),
    pytest.param(
        bytes.fromhex(
            "07"  # Subopcode
            "0000000000"  # ELT Test TAI Timestamp: TAI Seconds
            "40"  # ELT Test TAI Timestamp: Time Zone Offset
            "FF00"  # ELT Test TAI Timestamp: TAI-UTC Delta
            "07"  # ELT Test Execution Status
            "01"  # ELT Test Execution Result
            "FFFF"  # ELT Duration Result
        ),
        EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    2000,
                    1,
                    1,
                    0,
                    0,
                    0,
                    tzinfo=timezone(timedelta(hours=0)),
                ),
                tai_utc_delta=timedelta(seconds=0),
            ),
            execution_result=dict(
                lamp_fault=True,
                battery_fault=False,
                circuit_fault=False,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.UNKNOWN,
            duration_result=65535,
        ),
    ),
    pytest.param(
        bytes.fromhex(
            "07"  # Subopcode
            "2E48D52F00"  # ELT Test TAI Timestamp: TAI Seconds
            "48"  # ELT Test TAI Timestamp: Time Zone Offset
            "2401"  # ELT Test TAI Timestamp: TAI-UTC Delta
            "02"  # ELT Test Execution Status
            "02"  # ELT Test Execution Result
            "0000"  # ELT Duration Result
        ),
        EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    2025,
                    6,
                    6,
                    8,
                    32,
                    41,
                    tzinfo=timezone(timedelta(hours=2)),
                ),
                tai_utc_delta=timedelta(seconds=37),
            ),
            execution_result=dict(
                lamp_fault=False,
                battery_fault=True,
                circuit_fault=False,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.IN_PROGRESS,
            duration_result=0,
        ),
    ),
    pytest.param(
        bytes.fromhex(
            "07"  # Subopcode
            "0000000000"  # ELT Test TAI Timestamp: TAI Seconds
            "78"  # ELT Test TAI Timestamp: Time Zone Offset
            "2401"  # ELT Test TAI Timestamp: TAI-UTC Delta
            "03"  # ELT Test Execution Status
            "0F"  # ELT Test Execution Result
            "100E"  # ELT Duration Result
            "45000000"  # ELT Test Relative Timestamp (optional)
        ),
        EmergencyLightingTestSubOpcode.ELT_DURATION_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    2000,
                    1,
                    1,
                    13,
                    59,
                    23,
                    tzinfo=timezone(timedelta(hours=14)),
                ),
                tai_utc_delta=timedelta(seconds=37),
            ),
            execution_result=dict(
                lamp_fault=True,
                battery_fault=True,
                circuit_fault=True,
                battery_duration_fault=True,
            ),
            execution_status=TestExecutionStatus.POSTPONED,
            duration_result=3600,
            relative_timestamp=69,
        ),
    ),
    pytest.param(
        bytes.fromhex("0884FF"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_EXECUTION_TIMEOUT,
        ),
    ),
    pytest.param(
        bytes.fromhex("0885FF"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_EXECUTION_TIMEOUT,
        ),
    ),
    pytest.param(
        bytes.fromhex("0886FF"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_RETRY_PERIOD,
        ),
    ),
    pytest.param(
        bytes.fromhex("0887FF"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_RETRY_PERIOD,
        ),
    ),
    pytest.param(
        bytes.fromhex("0888FF"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_DELAY,
        ),
    ),
    pytest.param(
        bytes.fromhex("0889FF"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_DELAY,
        ),
    ),
    pytest.param(
        bytes.fromhex("088AFF"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_INTERVAL,
        ),
    ),
    pytest.param(
        bytes.fromhex("088BFF"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_INTERVAL,
        ),
    ),
    pytest.param(
        bytes.fromhex("0984FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_EXECUTION_TIMEOUT,
            elt_duration_test_execution_timeout=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0985FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_EXECUTION_TIMEOUT,
            elt_functional_test_execution_timeout=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0986FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_RETRY_PERIOD,
            elt_duration_test_retry_period=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0987FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_RETRY_PERIOD,
            elt_functional_test_retry_period=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0988FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_DELAY,
            elt_duration_test_backup_automatic_delay=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0989FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_DELAY,
            elt_functional_test_backup_automatic_delay=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("098AFF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_INTERVAL,
            elt_duration_test_backup_automatic_interval=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("098BFF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_INTERVAL,
            elt_functional_test_backup_automatic_interval=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A84FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_EXECUTION_TIMEOUT,
            elt_duration_test_execution_timeout=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A85FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_EXECUTION_TIMEOUT,
            elt_functional_test_execution_timeout=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A86FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_RETRY_PERIOD,
            elt_duration_test_retry_period=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A87FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_RETRY_PERIOD,
            elt_functional_test_retry_period=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A88FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_DELAY,
            elt_duration_test_backup_automatic_delay=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A89FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_DELAY,
            elt_functional_test_backup_automatic_delay=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A8AFF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_INTERVAL,
            elt_duration_test_backup_automatic_interval=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A8BFF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_INTERVAL,
            elt_functional_test_backup_automatic_interval=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B84FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_EXECUTION_TIMEOUT,
            elt_duration_test_execution_timeout=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B85FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_EXECUTION_TIMEOUT,
            elt_functional_test_execution_timeout=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B86FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_RETRY_PERIOD,
            elt_duration_test_retry_period=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B87FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_RETRY_PERIOD,
            elt_functional_test_retry_period=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B88FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_DELAY,
            elt_duration_test_backup_automatic_delay=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B89FF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_DELAY,
            elt_functional_test_backup_automatic_delay=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B8AFF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_DURATION_TEST_BACKUP_AUTOMATIC_INTERVAL,
            elt_duration_test_backup_automatic_interval=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B8BFF04030201"),
        EmergencyLightingTestSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=EmergencyLightingTestProperty.ELT_FUNCTIONAL_TEST_BACKUP_AUTOMATIC_INTERVAL,
            elt_functional_test_backup_automatic_interval=dict(seconds=0x01020304),
        ),
    ),
]


@pytest.mark.parametrize("encoded, subopcode, payload", valid)
def test_parse(encoded, subopcode, payload):
    decoded = EmergencyLightingTestParams.parse(encoded)
    assert decoded.subopcode == subopcode
    assert decoded.payload == payload


@pytest.mark.parametrize("encoded, subopcode, payload", valid)
def test_build(encoded, subopcode, payload):
    assert EmergencyLightingTestParams.build(dict(subopcode=subopcode, payload=payload)) == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2024  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import Construct, Int8ul, Int16ul, Int32ul, Struct, this

from bluetooth_mesh.messages.properties import PerceivedLightness, PropertyMixin, TimeSecond16
from bluetooth_mesh.messages.util import EnumAdapter, EnumSwitch as Switch, Opcode, SwitchStruct


class EmergencyLightingOpcode(IntEnum):
    SILVAIR_EL = 0xEA3601


class EmergencyLightingSubOpcode(IntEnum):
    EL_INHIBIT_ENTER = 0x00
    EL_INHIBIT_ENTER_UNACKNOWLEDGED = 0x01
    EL_INHIBIT_EXIT = 0x02
    EL_INHIBIT_EXIT_UNACKNOWLEDGED = 0x03
    EL_STATE_GET = 0x04
    EL_STATE_STATUS = 0x05
    EL_PROPERTY_GET = 0x06
    EL_PROPERTY_SET = 0x07
    EL_PROPERTY_SET_UNACKNOWLEDGED = 0x08
    EL_PROPERTY_STATUS = 0x09
    EL_LAMP_OPERATION_TIME_GET = 0x0A
    EL_LAMP_OPERATION_TIME_CLEAR = 0x0B
    EL_LAMP_OPERATION_TIME_CLEAR_UNACKNOWLEDGED = 0x0C
    EL_LAMP_OPERATION_TIME_STATUS = 0x0D
    EL_REST_ENTER = 0x0E
    EL_REST_ENTER_UNACKNOWLEDGED = 0x0F
    EL_REST_EXIT = 0x10
    EL_REST_EXIT_UNACKNOWLEDGED = 0x11


class ElState(IntEnum):
    TRANSITION_FROM_INHIBIT_TO_NORMAL = 0x00
    TRANSITION_FROM_FUNCTIONAL_TEST_TO_NORMAL = 0x01
    TRANSITION_FROM_DURATION_TEST_TO_NORMAL = 0x02
    NORMAL = 0x03
    TRANSITION_FROM_REST_TO_EMERGENCY = 0x04
    EMERGENCY = 0x05
    EXTENDED_EMERGENCY = 0x06
    TRANSITION_FROM_EMERGENCY_TO_REST = 0x07
    REST = 0x08
    TRANSITION_FROM_NORMAL_TO_INHIBIT = 0x09
    INHIBIT = 0x0A
    TRANSITION_FROM_NORMAL_TO_DURATION_TEST = 0x0B
    DURATION_TEST_IN_PROGRESS = 0x0C
    TRANSITION_FROM_NORMAL_TO_FUNCTIONAL_TEST = 0x0D
    FUNCTIONAL_TEST_IN_PROGRESS = 0x0E
    BATTERY_DISCHARGED = 0x0F
    UNKNOWN = 0x10


class EmergencyLightingProperty(IntEnum):
    EL_LIGHTNESS = 0xFF80
    EL_LIGHTNESS_RANGE_MIN = 0xFF81
    EL_LIGHTNESS_RANGE_MAX = 0xFF82
    EL_PROLONG_TIME = 0xFF83


EmergencyLightingPropertyDict = {
    EmergencyLightingProperty.EL_LIGHTNESS: PerceivedLightness,
    EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MIN: PerceivedLightness,
    EmergencyLightingProperty.EL_LIGHTNESS_RANGE_MAX: PerceivedLightness,
    EmergencyLightingProperty.EL_PROLONG_TIME: TimeSecond16,
}


class _EmergencyLightingProperty(PropertyMixin, Construct):
    ENUM = EmergencyLightingProperty
    DICT = EmergencyLightingPropertyDict

    subcon = Struct(
        "property_id" / EnumAdapter(Int16ul, EmergencyLightingProperty),
        Switch(this.id, EmergencyLightingPropertyDict),
    )

    def _parse(self, stream, context, path):
        msg = EmergencyLightingPropertyGet._parse(stream, context, path)
        return self._parse_property(msg, stream, context, path)

    def _build(self, obj, stream, context, path):
        EmergencyLightingPropertyGet._build(obj, stream, context, path)
        return self._build_property(obj, stream, context, path)


# fmt: off
EmergencyLightingInhibitEnter = Struct()

EmergencyLightingInhibitExit = Struct()

EmergencyLightingStateGet = Struct()

EmergencyLightingStateStatus = Struct(
    "el_state" / EnumAdapter(Int8ul, ElState),
)

EmergencyLightingPropertyGet = Struct(
    "property_id" / EnumAdapter(Int16ul, EmergencyLightingProperty),
)

EmergencyLightingPropertySet = _EmergencyLightingProperty()

EmergencyLightingPropertyStatus = _EmergencyLightingProperty()

EmergencyLightingLampOperationTimeGet = Struct()

EmergencyLightingLampOperationTimeClear = Struct()

EmergencyLightingLampOperationTimeStatus = Struct(
    "el_lamp_total_operation_time" / Int32ul,
    "el_lamp_emergency_time" / Int32ul,
)

EmergencyLightingRestEnter = Struct()

EmergencyLightingRestExit = Struct()

EmergencyLightingParams = SwitchStruct(
    "subopcode" / EnumAdapter(Int8ul, EmergencyLightingSubOpcode),
    "payload" / Switch(
        this.subopcode,
        {
            EmergencyLightingSubOpcode.EL_INHIBIT_ENTER: EmergencyLightingInhibitEnter,
            EmergencyLightingSubOpcode.EL_INHIBIT_ENTER_UNACKNOWLEDGED: EmergencyLightingInhibitEnter,
            EmergencyLightingSubOpcode.EL_INHIBIT_EXIT: EmergencyLightingInhibitExit,
            EmergencyLightingSubOpcode.EL_INHIBIT_EXIT_UNACKNOWLEDGED: EmergencyLightingInhibitExit,
            EmergencyLightingSubOpcode.EL_STATE_GET: EmergencyLightingStateGet,
            EmergencyLightingSubOpcode.EL_STATE_STATUS: EmergencyLightingStateStatus,
            EmergencyLightingSubOpcode.EL_PROPERTY_GET: EmergencyLightingPropertyGet,
            EmergencyLightingSubOpcode.EL_PROPERTY_SET: EmergencyLightingPropertySet,
            EmergencyLightingSubOpcode.EL_PROPERTY_SET_UNACKNOWLEDGED: EmergencyLightingPropertySet,
            EmergencyLightingSubOpcode.EL_PROPERTY_STATUS: EmergencyLightingPropertyStatus,
            EmergencyLightingSubOpcode.EL_LAMP_OPERATION_TIME_GET: EmergencyLightingLampOperationTimeGet,
            EmergencyLightingSubOpcode.EL_LAMP_OPERATION_TIME_CLEAR: EmergencyLightingLampOperationTimeClear,
            EmergencyLightingSubOpcode.EL_LAMP_OPERATION_TIME_CLEAR_UNACKNOWLEDGED: EmergencyLightingLampOperationTimeClear,
            EmergencyLightingSubOpcode.EL_LAMP_OPERATION_TIME_STATUS: EmergencyLightingLampOperationTimeStatus,
            EmergencyLightingSubOpcode.EL_REST_ENTER: EmergencyLightingRestEnter,
            EmergencyLightingSubOpcode.EL_REST_ENTER_UNACKNOWLEDGED: EmergencyLightingRestEnter,
            EmergencyLightingSubOpcode.EL_REST_EXIT: EmergencyLightingRestExit,
            EmergencyLightingSubOpcode.EL_REST_EXIT_UNACKNOWLEDGED: EmergencyLightingRestExit,
        }
    )
)

EmergencyLightingMessage = SwitchStruct(
    "opcode" / Opcode(EmergencyLightingOpcode),
    "params" / Switch(
        this.opcode,
        {
            EmergencyLightingOpcode.SILVAIR_EL: EmergencyLightingParams
        }
    )
)
# fmt: on
